import functools
import json
from typing import Any, Callable, Dict, List, Optional

from ..logging import log
from ..model.errors import HTTPError
from .coprocessor import Coprocessor, CoprocessorEvent
from .event_handler import EventHandler
from .executor import CoprocessorExecutor, RealCoProcessorExecutor
from .smartpipe import SmartPipe, SmartPipeEvent


def coprocessor_to_json(coprocessor: Coprocessor) -> Dict[str, Any]:
    return {"id": coprocessor.id, "type": coprocessor.type, "model": coprocessor.model}


def smart_pipe_to_json(smart_pipe: SmartPipe) -> Dict[str, Any]:
    return {
        "id": smart_pipe.id,
        "path": smart_pipe.path,
        "method": smart_pipe.method,
        "coprocessors": [
            coprocessor_to_json(coprocessor) for coprocessor in smart_pipe.coprocessors
        ],
    }


def smart_pipes_json(smart_pipes: List[SmartPipe]) -> Dict[str, Any]:
    return {
        "type": "smart_pipes",
        "payload": [smart_pipe_to_json(smart_pipe) for smart_pipe in smart_pipes],
    }


class Context:
    def __init__(
        self,
        smart_pipes: Optional[List[SmartPipe]] = None,
        coprocessors: Optional[List[Coprocessor]] = None,
    ) -> None:
        self.actual_smart_pipe_index = -1
        self.event_handler = EventHandler()
        self.coprocessor_executor: CoprocessorExecutor = RealCoProcessorExecutor(
            self.event_handler
        )

        if smart_pipes is None:
            self.smart_pipes = []
        else:
            self.smart_pipes = smart_pipes

        if coprocessors is None:
            self.coprocessors = []
        else:
            self.coprocessors = coprocessors

    def active_smart_pipe(self, id: str) -> None:
        for i, smart_pipe in enumerate(self.smart_pipes):
            if smart_pipe.id == id:
                self.actual_smart_pipe_index = i
                break

    def set_event(self, callback: Callable[[Dict[str, Any]], None]) -> None:
        self.event_handler.set_event(callback)

    def add_event(self, event: SmartPipeEvent) -> None:
        self.event_handler.add_event(event)

    def update_event(self, event: SmartPipeEvent) -> None:
        self.event_handler.update_event(event)

    def coprocessor_event(self, coprocessor_event: CoprocessorEvent) -> None:
        self.event_handler.coprocessor_event(coprocessor_event)

    def get_actual_smart_pipe(self) -> Optional[SmartPipe]:
        if self.actual_smart_pipe_index == -1:
            return None

        return self.smart_pipes[self.actual_smart_pipe_index]

    def add_smart_pipe(self, smart_pipe: SmartPipe) -> None:
        if len(self.smart_pipes) == 1 and self.smart_pipes[0].id == "temporal":
            self.smart_pipes[0] = smart_pipe
        else:
            self.actual_smart_pipe_index = len(self.smart_pipes)
            self.smart_pipes.append(smart_pipe)

        log.info(f"🧠 Smart Pipe: {smart_pipe.id}, Path: {smart_pipe.path}")
        self.event_handler.on_change(smart_pipes_json(self.smart_pipes))

    def add_coprocessor(self, coprocessor: Coprocessor) -> None:
        log.info(f"⌛️ Coprocessor {coprocessor.id} of type: {coprocessor.type}")
        self.coprocessors.append(coprocessor)

    def get_coprocessor(self, id: str) -> Optional[Coprocessor]:
        for c in self.coprocessors:
            if c.id == id:
                return c

        return None

    def assign_to_active_smart_pipe(self, coprocessor: Coprocessor) -> None:
        self.smart_pipes[self.actual_smart_pipe_index].add_coprocessor(coprocessor)
        smart_pipe = context.get_actual_smart_pipe()
        if smart_pipe is not None:
            log.info(f"⌛️ Assign Coprocessor {coprocessor.id} to Smart Pipe: {smart_pipe.id}")
        else:
            log.info(
                f"🔥 Actual Smart Pipe is None, can't assign \
                    Coprocessor {coprocessor.id} to Smart Pipe"
            )
        self.event_handler.on_change(smart_pipes_json(self.smart_pipes))

    def set_executor(self, executor: CoprocessorExecutor) -> None:
        self.coprocessor_executor = executor


context = Context()


def smartpipe(
    path: str,
    method: str,
    id: str,
    parameters: Optional[List[str]] = None,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Any, Any], Any]:
    def decorator_smartpipes(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            log.info(f"SmartPipe Path: {path}, Method: {method}, ID: {id}")
            context.active_smart_pipe(id)

            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = SmartPipeEvent(smart_pipe_id=id, input=args_json)
            context.add_event(event)
            result = None

            try:
                result = func(*args, **kwargs)
                event.set_ouput(result)
            except HTTPError as err:
                log.error(f"Smart Pipe error: {err}")
                event.set_error(err)

            context.update_event(event)
            return result

        smart_pipe = SmartPipe(wrapper, path, method, id, parameters)
        context.add_smart_pipe(smart_pipe)
        return wrapper

    if not _func:
        return decorator_smartpipes  # type: ignore
    else:
        return decorator_smartpipes(_func)


def import_coprocessor(
    _func: Optional[Callable[[Any], Any]], coprocessor: Coprocessor
) -> Callable[[Any, Any], Any]:
    def decorator_coprocessor(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
            args_json = json.dumps(args_str)

            event = CoprocessorEvent(coprocessor.id, args_json)
            context.coprocessor_event(event)

            result = coprocessor.process(*args, **kwargs)

            event.set_ouput(result)
            context.coprocessor_event(event)

            return func(result)

        return wrapper

    context.add_coprocessor(coprocessor)

    if not _func:
        return decorator_coprocessor  # type: ignore
    else:
        return decorator_coprocessor(_func)


def coprocessor(
    type: str,
    id: str,
    model: Optional[str] = None,
    _func: Optional[Callable[[Any], Any]] = None,
) -> Callable[[Any, Any], Any]:
    def decorator_coprocessor(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:

        coprocessor = Coprocessor(func=func, type=type, model=model, id=id)
        context.add_coprocessor(coprocessor)

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            context.assign_to_active_smart_pipe(coprocessor)

            return context.coprocessor_executor.execute(coprocessor, *args, **kwargs)

        return wrapper

    if not _func:
        return decorator_coprocessor  # type: ignore
    else:
        return decorator_coprocessor(_func)



import os
from typing import Any, Callable, Dict, List, Optional, Tuple

from langchain.chains import ConversationalRetrievalChain
from langchain.document_loaders import PyPDFLoader
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.llms import OpenAI
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import Pinecone
import pinecone

from ...configuration import config
from ...logging import log
from ...model.errors import SeaplaneError

_PINECONE_INDEX_NAME = "PINECONE_INDEX_ENV"
_PINECONE_API_KEY_NAME = "PINECONE_API_KEY"
_PINECONE_API_ENV_NAME = "PINECONE_API_ENV"


def _check_pinecone_api_key() -> None:
    if config._api_keys is None:
        raise SeaplaneError(
            f"Pinecone API Key `{_PINECONE_API_KEY_NAME}` is not set,\
                  use `sea.config.set_api_keys`."
        )
    elif not config._api_keys.get(_PINECONE_API_KEY_NAME, None) or not config._api_keys.get(
        _PINECONE_API_ENV_NAME, None
    ):
        raise SeaplaneError(
            f"Pinecone API Key `{_PINECONE_API_KEY_NAME}` or `{_PINECONE_API_ENV_NAME}`\
                  is not set, use `sea.config.set_api_keys`."
        )


def get_index() -> str:
    index = os.getenv(_PINECONE_INDEX_NAME)

    if index is None:
        return ""

    return index


class Store:
    def __init__(self) -> None:
        _check_pinecone_api_key()

        self.index = get_index()
        self.chat_history_file: Dict[str, List[Tuple[str, str]]] = {}
        self.text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        self.embeddings = OpenAIEmbeddings(  # type: ignore
            openai_api_key=config._api_keys.get("OPENAI_API_KEY", None)
        )
        pinecone.init(
            api_key=config._api_keys.get(_PINECONE_API_KEY_NAME, None),
            environment=config._api_keys.get(_PINECONE_API_ENV_NAME, None),
        )

    def save(self, file_name: str, file_url: str) -> None:
        loader = PyPDFLoader(file_url)
        document = loader.load()
        texts = self.text_splitter.split_documents(document)

        log.info(f"⏳ Saving file {file_name}")
        Pinecone.from_texts(
            [chunk.page_content for chunk in texts],
            self.embeddings,
            index_name=self.index,
            namespace=file_name,
        )

    def query(self, file_name: str, query: str) -> Dict[str, Any]:
        vectorstore = Pinecone.from_existing_index(
            index_name=self.index, embedding=self.embeddings, namespace=file_name
        )
        qa = ConversationalRetrievalChain.from_llm(
            llm=OpenAI(  # type: ignore
                model="gpt-3.5-turbo",
                temperature=0.7,
                openai_api_key=config._api_keys.get("OPENAI_API_KEY", None),
            ),
            retriever=vectorstore.as_retriever(),
            return_source_documents=True,
        )

        history = self.chat_history_file.get(file_name, None)
        if history is None:
            history = []
            self.chat_history_file[file_name] = []

        result = qa({"question": query, "chat_history": history})
        self.chat_history_file[file_name].append((query, result["answer"]))

        return {"answer": result["answer"], "history": history}


class PineconeCoprocessor:
    def __init__(self, func: Callable[[Any], Any], id: str, model: Optional[str]) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = "pinecone"
        self.model = model
        self.id = id

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        if self.type == "pinecone":
            log.info("Accessing Vector DB coprocessor...")
            self.args = self.args + (Store(),)

            return self.func(*self.args, **self.kwargs)

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")



import json
import os
import shutil

from typing import Any, Dict, List

from .decorators import context
from .coprocessor import Coprocessor
from ..logging import log
from .build import build
from ..configuration import config

import requests

from ..api.api_http import headers
from ..api.api_request import provision_req
from ..util import unwrap


def create_coprocessor_docker_file(coprocessor: Coprocessor) -> None:
    docker_file = f"""FROM python:3.10

ENV SMARTPIPES_PRODUCTION True
ENV COPROCESSOR_ID {coprocessor.id}
WORKDIR /app
COPY . . 

RUN pip install --no-cache-dir -r requirements.txt
ENTRYPOINT ["python", "demo.py"]
    """

    if not os.path.exists(f"build/{coprocessor.id}"):
        os.makedirs(f"build/{coprocessor.id}")

    with open(f"build/{coprocessor.id}/Dockerfile", "w") as file:
        file.write(docker_file)        
        
    os.system(f"docker buildx build --platform linux/arm64,linux/amd64 -t us-central1-docker.pkg.dev/artifacts-356722/demo/coprocessors/{coprocessor.id}:latest  build/{coprocessor.id}  --push")


def create_http_api_entry_point() -> None:
    docker_file = """FROM python:3.10

ENV SMARTPIPES_PRODUCTION True
ENV PORT 5000

WORKDIR /app
COPY . . 

RUN pip install --no-cache-dir -r requirements.txt
RUN pip install gunicorn


EXPOSE 5000

CMD gunicorn --bind 0.0.0.0:${PORT} --workers 5 --timeout 300 demo:sea
    """

    if not os.path.exists(f"build/http"):
        os.makedirs(f"build/http")

    with open("build/http/Dockerfile", "w") as file:
        file.write(docker_file)


def create_carrier_workload_file(
    smart_pipe_id: str, coprocessor: Coprocessor, next_coprocessors: List[str]
) -> None:
    output = None

    if len(next_coprocessors) > 1:
        output = {
            "broker": {
                "outputs": map(
                    lambda c_id: {"carrier": {"subject": f"{smart_pipe_id}.{c_id}"}},
                    next_coprocessors,
                )
            }
        }
    elif len(next_coprocessors) == 1:
        output = {
            "label": "carrier_out",
            "carrier": {"subject": f"{smart_pipe_id}.{next_coprocessors[0]}"},
        }

    workload = {
        "tenant": "tnt-0oqdcinn0pubb000tq8ng6qoos",
        "id": coprocessor.id,
        "input": {
            "label": "carrier_in",
            "carrier": {
                "subject": f"{smart_pipe_id}.{coprocessor.id}",
                "deliver": "all",
                "queue": coprocessor.id,
            },
        },
        "processor": {
            "docker": {
                "image": f"us-central1-docker.pkg.dev/artifacts-356722/demo/coprocessors/{coprocessor.id}:latest",
                "args": [],
            }
        },
        "output": output,
    }

    if not os.path.exists(f"build/{coprocessor.id}"):
        os.makedirs(f"build/{coprocessor.id}")

    with open(f"build/{coprocessor.id}/workload.json", "w") as file:
        json.dump(workload, file, indent=2)
        log.debug(f"Created {coprocessor.id} workload")

    return workload


def copy_project_into_coprocessors(coprocessor: Coprocessor) -> None:
    source_folder = "."
    destination_folder = f"build/{coprocessor.id}"

    if not os.path.exists(f"build/{coprocessor.id}"):
        os.makedirs(f"build/{coprocessor.id}")

    for item in os.listdir(source_folder):        
        if os.path.isdir(item) and item == "build":
            continue  # Skip the "build" folder

        destination_path = os.path.join(destination_folder, item)
        shutil.copy2(item, destination_path)


def create_stream(name: str) -> Any:
    log.debug(f"Creating stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"
    req = provision_req(config._token_api)

    payload = {
        "message_ttl": 3600,
        "max_messages": 1000000,
        "max_size": 100000000,
        "replicas": 3,
        "allow_locations": ["region/xn"],
        "deny_locations": ["country/nl"],
        "wait_for_ack": True,
        "ack_timeout": 5,
        "max_delivery_attempts": 5,
        "max_delivery_time": 60,
        "dead_letter_sink": "dead-letter-sink"
    }

    return unwrap(
        req(
            lambda access_token: requests.put(
                url,
                json=payload,
                headers=headers(access_token),
            )
        )
    )


def delete_stream(name: str) -> Any:
    log.debug(f"Deleting stream: {name}")
    url = f"{config.carrier_endpoint}/stream/{name}"
    req = provision_req(config._token_api)

    return unwrap(
        req(
            lambda access_token: requests.delete(
                url,
                headers=headers(access_token),
            )
        )
    )


def create_flow(name: str, workload: Dict[str, Any]) -> Any:
    log.debug(f"Creating flow: {name}")
    url = f"{config.carrier_endpoint}/flow/{name}"
    req = provision_req(config._token_api)

    return unwrap(
        req(
            lambda access_token: requests.put(
                url,
                json=workload,
                headers=headers(access_token),
            )
        )
    )


def delete_flow(name: str) -> Any:
    log.debug(f"Deleting flow: {name}")

    url = f"{config.carrier_endpoint}/flow/{name}"
    req = provision_req(config._token_api)

    return unwrap(
        req(
            lambda access_token: requests.delete(
                url,                
                headers=headers(access_token),
            )
        )
    )


def deploy() -> None:
    schema = build()

    for sm in context.smart_pipes:
        delete_stream(sm.id)
        create_stream(sm.id)

        for c in sm.coprocessors:
            delete_flow(c.id)

            copy_project_into_coprocessors(c)
            create_coprocessor_docker_file(c)
            next_coprocessors = schema["smartpipes"][sm.id]["io"].get(c.id, None)

            if next_coprocessors is None:
                next_coprocessors = []

            workload = create_carrier_workload_file(sm.id, c, next_coprocessors)            

            create_flow(c.id, workload)

            log.info(f"Deploy for coprocessor {c.id} done")

    create_http_api_entry_point()
    log.info(f"Build for coprocessor {c.id} done")



from .api.compute_api import ComputeAPI
from .api.lock_api import LockAPI
from .api.metadata_api import MetadataAPI
from .api.restrict_api import RestrictAPI
from .api.sql_api import GlobalSQL
from .api.token_api import TokenAPI
from .configuration import Configuration, config
from .logging import log
from .smartpipes import (
    Coprocessor,
    SmartPipe,
    build,
    context,
    coprocessor,
    smartpipe,
    start,
    deploy,
)


class Seaplane:
    @property
    def config(self) -> Configuration:
        return config

    @property
    def auth(self) -> TokenAPI:
        return config._token_api

    @property
    def metadata(self) -> MetadataAPI:
        return MetadataAPI(config)

    @property
    def locks(self) -> LockAPI:
        return LockAPI(config)

    @property
    def restrict(self) -> RestrictAPI:
        return RestrictAPI(config)

    @property
    def compute(self) -> ComputeAPI:
        return ComputeAPI(config)

    @property
    def global_sql(self) -> GlobalSQL:
        return GlobalSQL(config)


sea = Seaplane()

__all__ = [
    "SmartPipe",
    "Coprocessor",
    "coprocessor",
    "smartpipe",
    "start",
    "log",
    "context",
    "build",
    "deploy",
]



"""process module hooks up to input/output FIFOs.

The default FIFOs are input.fifo (read from) and output.fifo (write
to). These can be overridden with `SEAPLANE_INPUT_FIFO` and
`SEAPLANE_OUTPUT_FIFO` environment variables respectively.

Usage::

    from seaplane.carrier import processor

    # Reverse all messages.
    while True:
        msg = processor.read()  # Read an entire message (bytes).
        processor.write(msg[::-1])  # Write the reversed bytes.

"""
import os
import logging
import threading
import time


class ProcessorIO:
    """ProcessorIO handles IO to a stdpipe with a framed transport.

    ProcessorIO provides two thread-safe methods, `read` and
    `write`. The frame format is a 4 byte header (unsigned 32 bit int,
    in big-endian order) indicating the frame size, followed by the
    frame itself.
    """

    def __init__(self, input_fifo_path, output_fifo_path):
        self._input_fifo_path = input_fifo_path
        self._output_fifo_path = output_fifo_path
        self._input_fifo = None
        self._output_fifo = None
        self._read_lock = threading.Lock()
        self._write_lock = threading.Lock()

    def read(self):
        """Read an entire framed message."""
        with self._read_lock:
            # Read frame header.
            header = self._input_fifo.read(4)
            frame_size = int.from_bytes(header, "big", signed=False)
            return self._input_fifo.read(frame_size)

    def write(self, msg):
        """Frame and write a message."""
        with self._write_lock:
            frame_size = len(msg)
            header = frame_size.to_bytes(4, "big")
            self._output_fifo.write(header + msg)
            self._output_fifo.flush()

    def start(self):
        logging.info("opening input {}".format(self._input_fifo_path))
        logging.info("opening output {}".format(self._output_fifo_path))
        for _ in range(10):
            try:
                self._input_fifo = open(self._input_fifo_path, "rb")
                self._output_fifo = open(self._output_fifo_path, "wb")
            except FileNotFoundError as exc:
                logging.error(exc)
                time.sleep(2)
            else:
                return
        raise FileNotFoundError("input/output")


processor = ProcessorIO(
    os.getenv("SEAPLANE_INPUT_FIFO", "input.fifo"),
    os.getenv("SEAPLANE_OUTPUT_FIFO", "output.fifo"),
)



from typing import Any, Dict, List, NamedTuple, Optional

from ...util.base64url import base64url_decode_to_bytes


class Name(NamedTuple):
    """
    bytes Name of a lock.
    """

    name: bytes


class NameString(Name):
    """
    String Name of a lock.

    Attributes
        ----------
        name : bytes
            name of a lock in bytes.
        name_str: str
            value of a lock in string
    """

    name_str: str

    def __new__(cls, name: str, encoding: str = "utf-8", **kwargs):  # type: ignore
        return super().__new__(cls, name.encode(encoding), **kwargs)

    def __init__(self, name: str):
        self.name_str = name


class LockInfo(NamedTuple):
    """
    Lock information class.

    Attributes
        ----------
        ttl : int
            This is the requested time to live, in seconds.
        client_id: str
            Client-chosen identifier stored with the lock for informational purposes.
        ip: str
            Lock IP

    """

    ttl: int
    client_id: str
    ip: str


class Lock(NamedTuple):
    """
    Lock class.

    Attributes
        ----------
        id : str
            Lock ID.
        name: Name
            Lock name.
        info: LockInfo
            Additional information of a lock.
    """

    id: str
    name: Name
    info: LockInfo


class LockPage(NamedTuple):
    """
    Lock Page class.

    It contains a paginated list of locks, next_lock is used for the next call.

    Attributes
        ----------
        locks : List[Lock]
            list of locks.
        next_lock: Optional[Name]
            If next_lock is non-null and you want to get the next LockPage,
            you can repeat a query using next_lock to continue getting LockPage.
    """

    locks: List[Lock]
    next_lock: Optional[Name]


class HeldLock(NamedTuple):
    """
    HeldLock class.

    Attributes
        ----------
        id : str
            The ID of the lock, to be used in release and renew calls.
        sequencer: int
            A sequencer representing the number of times the lock has gone
            from held to free, can be used to safely coordinate
            access with external resources.
    """

    id: str
    sequencer: int


def to_lock_page(lock_range: Dict[str, Any]) -> LockPage:
    return LockPage(
        locks=[to_lock(lock) for lock in lock_range["locks"]],
        next_lock=_to_name(lock_range["next"]),
    )


def _to_name(name: Optional[str]) -> Optional[Name]:
    if name is None:
        return None

    return Name(base64url_decode_to_bytes(name))


def to_lock(lock: Dict[str, Any]) -> Lock:
    lock_name = _to_name(lock["name"])
    lock_name = Name(b"") if lock_name is None else lock_name
    return Lock(id=lock["id"], name=lock_name, info=to_lock_info(lock["info"]))


def to_lock_info(lock_info: Dict[str, Any]) -> LockInfo:
    return LockInfo(ttl=lock_info["ttl"], client_id=lock_info["client-id"], ip=lock_info["ip"])


def to_held_lock(lock: Dict[str, Any]) -> HeldLock:
    return HeldLock(id=lock["id"], sequencer=lock["sequencer"])



from typing import Any, Callable, Dict, Optional, Tuple

import requests

from ...api.api_http import headers
from ...api.api_request import provision_req
from ...configuration import config
from ...logging import log
from ...model.errors import SeaplaneError
from ...util import unwrap

SEAPLANE_API_KEY_NAME = "SEAPLANE_API_KEY"


def _check_seaplane_api_key() -> None:
    if config.seaplane_api_key is None:
        raise SeaplaneError(
            f"Seaplane API Key `{SEAPLANE_API_KEY_NAME}` is not set,\
                  use `sea.config.set_api_key`."
        )


def bloom() -> Callable[[Dict[str, Any]], Any]:
    def model(input: Dict[str, Any]) -> Any:
        _check_seaplane_api_key()

        url = f"{config.substation_endpoint}/completions"
        req = provision_req(config._token_api)

        payload = {"prompt": input}

        return unwrap(
            req(
                lambda access_token: requests.post(
                    url,
                    json=payload,
                    headers=headers(access_token),
                )
            )
        )

    return model


class Bloom:
    def __init__(self, func: Callable[[Any], Any], id: str, model: Optional[str]) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = "inference"
        self.model = model
        self.id = id

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        if self.model == "bloom":
            log.info("Processing Bloom Model...")
            self.args = self.args + (bloom(),)

            return self.func(*self.args, **self.kwargs)

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")



from .compute import *  # noqa
from .locks import *  # noqa
from .metadata import *  # noqa
from .provider import *  # noqa
from .region import *  # noqa
from .restrict import *  # noqa
from .sql import CreatedDatabase, to_created_database, to_list_databases

__all__ = ("CreatedDatabase", "to_created_database", "to_list_databases")



import requests

from ..configuration import Configuration, config
from ..model.compute import Formation, FormationPage, to_formation, to_formation_page
from ..util import unwrap
from .api_http import headers
from .api_request import provision_req


class ComputeAPI:
    """
    Class for handle Compute API calls.
    """

    def __init__(self, configuration: Configuration = config) -> None:
        self.url = f"{configuration.compute_endpoint}/formations"
        self.req = provision_req(configuration._token_api)

    def create(self, formation: Formation) -> None:
        """Create a new Formation and returns the IDs of the created Formation.

        Parameters
        ----------
        formation : Formation
            formation with flights to be created and deployed.
        """

        _payload = {
            "oid": formation.oid,
            "name": formation.name,
            "url": formation.url,
            "flights": [
                {"name": flight.name, "oid": flight.oid, "image": flight.image}
                for flight in formation.flights
            ],
            "gateway-flight": formation.gateway_flight,
        }

        unwrap(
            self.req(
                lambda access_token: requests.post(
                    self.url,
                    json=_payload,
                    headers=headers(access_token),
                )
            )
        )

    def get(self, formation_id: str) -> Formation:
        """Returns the Formation associated with the created Formation OID.

        Parameters
        ----------
        formation_id : str
            The id from a Formation previously created.

        Returns
        -------
        Formation
            Returns Formation if successful or it will raise an HTTPError otherwise.
        """
        url = f"{self.url}/{formation_id}"

        return unwrap(
            self.req(lambda access_token: requests.get(url, headers=headers(access_token))).map(
                lambda formation: to_formation(formation)
            )
        )

    def delete(self, formation_id: str) -> None:
        """Deletes the Formation from a given oid.

        Parameters
        ----------
        formation_id : str
            The oid from a Formation previously created.
        """
        url = f"{self.url}/{formation_id}"

        unwrap(self.req(lambda access_token: requests.delete(url, headers=headers(access_token))))

    def get_page(
        self,
    ) -> FormationPage:
        """Returns a list of all the Formations you have access to

        Returns
        -------
        List[Formations]
            Returns Formations and meta info for next and previous pages if successful
            or it will raise an HTTPError otherwise.
        """

        return unwrap(
            self.req(
                lambda access_token: requests.get(self.url, headers=headers(access_token))
            ).map(lambda formation_page: to_formation_page(formation_page))
        )



import functools
import os
import sys
import json
from typing import Any, Optional

from flask import Flask, request
from flask_cors import CORS
from flask_socketio import SocketIO, emit

from ..configuration import config
from ..logging import log
from .decorators import context, smart_pipes_json
from .smartpipe import SmartPipe
from ..model.errors import SeaplaneError
from ..carrier import processor
from .build import build
from .deploy import deploy

SMARTPIPES_CORS = list(os.getenv("SMARTPIPES_CORS", "http://localhost:3000").split(" "))
AUTH_TOKEN = os.getenv("SMARTPIPES_AUTH_TOKEN")

app = Flask(__name__)
sio = SocketIO(app, cors_allowed_origins=SMARTPIPES_CORS, async_mode="threading")
CORS(app, origins=SMARTPIPES_CORS)


@sio.on("message")  # type: ignore
def handle_message(data: Any) -> None:
    ...


@sio.on("connect")  # type: ignore
def on_connect() -> None:
    emit("message", smart_pipes_json(context.smart_pipes))


def send_something(data: Any) -> None:
    emit("message", data, sid="sp", namespace="", broadcast=True)


def authenticate_token() -> bool:
    token = request.headers.get("Authorization")
    if token and token.startswith("Bearer "):

        if token == f"Bearer {AUTH_TOKEN}":
            return True

    return False


@app.before_request
def before_request() -> Any:
    if request.path == "/healthz":
        return

    if os.getenv("SMARTPIPES_AUTH_TOKEN") is not None:
        if not authenticate_token():
            return {"message": "Unauthorized"}, 401


def https_api_start() -> Flask:
    log.debug(f"CORS enabled: {SMARTPIPES_CORS}")
    context.set_event(lambda data: send_something(data))

    smart_pipes = context.smart_pipes

    for smart_pipe in smart_pipes:

        def endpoint_func(pipe: SmartPipe = smart_pipe) -> Any:
            data = request.get_json()
            result = pipe.func(data)
            return result

        endpoint = functools.partial(endpoint_func, pipe=smart_pipe)
        app.add_url_rule(smart_pipe.path, smart_pipe.id, endpoint, methods=[smart_pipe.method])

    def health() -> str:
        emit("message", {"data": "test"}, sid="lol", namespace="", broadcast=True)
        return "Seaplane SmartPipes Demo"

    app.add_url_rule("/", "healthz", health, methods=["GET"])

    if not config.is_production():
        log.info("🚀 Smart Pipes DEVELOPMENT MODE")
        sio.run(app, debug=False, port=1337)
    else:
        log.info("🚀 Smart Pipes PRODUCTION MODE")

    return app


def start_coprocessor(coprocessor_id: str) -> None:
    coprocessor = context.get_coprocessor(coprocessor_id)

    if not coprocessor:
        raise SeaplaneError(
            f"Coprocessor {coprocessor_id} not found, \
                            make sure the Coprocessor ID is correct."
        )

    processor.start()

    while True:
        try: 
            log.info(f" Coprocessor {coprocessor.id}  waiting for getting data...")
            message = json.loads(processor.read())
            log.debug(f" Message recieved: {message}")
                    
            message["output"] = coprocessor.process(message["input"])

            log.debug(f" Coprocessor Result: {message}")
                        

            next_message = { "input": message["output"] }
            processor.write(str(json.dumps(next_message)).encode())
        except Exception as ex:
            log.error(f"Error running Coprocessor: {str(ex)}")



def start() -> Optional[Flask]:
    if len(sys.argv) > 1:
        command = sys.argv[1]

        if command == "build":
            build()
            return None
        elif command == "deploy":
            deploy()
            return None

    coprocessor_id = os.getenv("COPROCESSOR_ID")

    if not coprocessor_id:
        log.info("Starting API Entry Point...")
        return https_api_start()
    else:
        log.info(f"Starting Coprocessor {coprocessor_id} ...")
        start_coprocessor(coprocessor_id)
        return None



import json
from typing import Any

from ..logging import log
from ..model.errors import HTTPError
from .coprocessor import Coprocessor, CoprocessorEvent
from .event_handler import EventHandler


class CoprocessorExecutor:
    def execute(self, coprocessor: Coprocessor, *args: Any, **kwargs: Any) -> Any:
        pass


class RealCoProcessorExecutor(CoprocessorExecutor):
    def __init__(self, event_handler: EventHandler) -> None:
        self.event_handler = event_handler

    def execute(self, coprocessor: Coprocessor, *args: Any, **kwargs: Any) -> Any:
        args_str = tuple(arg.decode() if isinstance(arg, bytes) else arg for arg in args)
        args_json = json.dumps(args_str)

        event = CoprocessorEvent(coprocessor.id, args_json)
        self.event_handler.coprocessor_event(event)
        result = None

        try:
            result = coprocessor.process(*args, **kwargs)
            event.set_ouput(result)
        except HTTPError as err:
            log.error(f"Coprocessor HTTPError: {err}")
            event.set_error(err)
        except Exception as e:
            log.error(f"Coprocessor error: {e}")
            event.set_error(e)

        self.event_handler.coprocessor_event(event)

        if event.error is not None:
            raise event.error

        return result


class SchemaExecutor(CoprocessorExecutor):
    def __init__(self) -> None:
        ...

    def execute(self, coprocessor: Coprocessor, *args: Any, **kwargs: Any) -> Any:
        arguments = []
        for arg in args:
            arguments.append(arg)

        coprocessor.called_from(arguments)
        return coprocessor.id



import json
import os
import shutil

from typing import Any, Dict, List

from .decorators import context
from .executor import SchemaExecutor
from .coprocessor import Coprocessor
from .smartpipe import SmartPipe
from ..logging import log


def persist_schema(schema: Dict[str, Any]) -> None:
    if not os.path.exists("build"):
        os.makedirs("build")

    file_path = os.path.join("build", "workload.json")

    with open(file_path, "w") as file:
        json.dump(schema, file, indent=2)


def build() -> Dict[str, Any]:
    schema: Dict[str, Any] = {"smartpipes": {}}

    context.set_executor(SchemaExecutor())
    for sm in context.smart_pipes:
        sm.func("entry_point")

    for sm in context.smart_pipes:
        smartpipe: Dict[str, Any] = {
            "id": sm.id,
            "entry_point": {"type": "API", "path": sm.path, "method": sm.method},
            "coprocessors": [],
            "io": {},
        }

        for c in sm.coprocessors:
            coprocessor = {"id": c.id, "name": c.name, "type": c.type, "model": c.model}

            for source in c.sources:
                if not smartpipe["io"].get(source, None):
                    smartpipe["io"][source] = [c.id]
                else:
                    smartpipe["io"][source].append(c.id)

            smartpipe["coprocessors"].append(coprocessor)

        schema["smartpipes"][sm.id] = smartpipe

    persist_schema(schema)
    log.debug("Created Smart Pipes configuration")

    return schema



from enum import Enum
from typing import Any, Dict, List, NamedTuple, Optional

from ..errors import SeaplaneError
from ..metadata import Key, _to_key
from ..provider import Provider, to_providers
from ..region import Region, to_regions


class SeaplaneApi(Enum):
    """
    API enum for which data placement can be restricted.


    Attributes
        ----------
        locks: str
            Locks API data placement that can be restricted.
        metadata: str
            Metadata API data placement that can be restricted.
    """

    locks = "Locks"
    metadata = "Config"

    def __str__(self) -> str:
        return str(self.value)


class RestrictionState(Enum):
    """
    State of a restriction.


    Attributes
        ----------
        enforced: str
            Restriction is in effect.
        pending: str
            Restriction is being applied.
    """

    enforced = "Enforced"
    pending = "Pending"

    def __str__(self) -> str:
        return str(self.value)


class RestrictionDetails(NamedTuple):
    """
    Allow / deny lists of cloud providers and geographic regions
    to be associated with a set of records.

    Attributes
        ----------
        regions_allowed: str
            List of Regions which are allowed.
        regions_denied: str
            List of Regions which are denied.
        providers_allowed: str
            List of Providers which are allowed.
        providers_denied: str
            List of Providers which are denied.
    """

    regions_allowed: List[Region]
    regions_denied: List[Region]
    providers_allowed: List[Provider]
    providers_denied: List[Provider]


class Restriction(NamedTuple):
    """
    The provider and region restrictions information.

    Attributes
        ----------
        api: SeaplaneApi
            An API for which data placement can be restricted.
        directory: bytes
            Key used to select a key-value pair or a directory
            in configuration API.
        details: RestrictionDetails
            Allow / deny lists of cloud providers and geographic
            regions to be associated with a set of records.
        state: RestrictionState
            State of a restriction.
    """

    api: SeaplaneApi
    directory: Key
    details: RestrictionDetails
    state: RestrictionState


class RestrictionPage(NamedTuple):
    """
    Restriction Page class.

    It contains a paginated list of restrictions, next_key is used for the next call.

    Attributes
        ----------
        restrictions : List[Restriction]
            list of restrictions.
        next_key: Optional[Name]
            If next_key is non-null and you want to get the next RestrictionPage,
            you can repeat a query using next_key to continue getting RestrictionPage.
    """

    restrictions: List[Restriction]
    next_api: Optional[SeaplaneApi]
    next_key: Optional[Key]


def to_restriction(restriction: Dict[str, Any]) -> Restriction:
    key = _to_key(restriction["directory"])
    if key is None:
        raise SeaplaneError("Directory must not be null")

    return Restriction(
        api=SeaplaneApi(restriction["api"]),
        directory=key,
        details=_to_restriction_details(restriction["details"]),
        state=RestrictionState(restriction["state"]),
    )


def _to_restriction_details(restriction: Dict[str, Any]) -> RestrictionDetails:
    return RestrictionDetails(
        providers_allowed=to_providers(restriction["providers_allowed"]),
        providers_denied=to_providers(restriction["providers_denied"]),
        regions_allowed=to_regions(restriction["regions_allowed"]),
        regions_denied=to_regions(restriction["regions_denied"]),
    )


def _to_seaplane_api(api: Optional[str]) -> Optional[SeaplaneApi]:
    if api is None:
        return None

    return SeaplaneApi(api.capitalize())


def to_restriction_page(restriction_page: Dict[str, Any]) -> RestrictionPage:
    return RestrictionPage(
        restrictions=[
            to_restriction(restriction) for restriction in restriction_page["restrictions"]
        ],
        next_api=_to_seaplane_api(restriction_page["next_api"]),
        next_key=_to_key(restriction_page["next_key"]),
    )



from .build import build
from .coprocessor import Coprocessor
from .decorators import context, coprocessor, smartpipe
from .smartapi import start
from .smartpipe import SmartPipe
from .deploy import deploy

__all__ = (
    "Coprocessor",
    "context",
    "coprocessor",
    "smartpipe",
    "start",
    "SmartPipe",
    "build",
    "deploy",
)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['seaplane',
 'seaplane.api',
 'seaplane.carrier',
 'seaplane.logging',
 'seaplane.model',
 'seaplane.model.compute',
 'seaplane.model.locks',
 'seaplane.model.metadata',
 'seaplane.model.restrict',
 'seaplane.model.sql',
 'seaplane.smartpipes',
 'seaplane.smartpipes.coprocessors',
 'seaplane.util']

package_data = \
{'': ['*'],
 'seaplane.smartpipes': ['website/.gitignore',
                         'website/.gitignore',
                         'website/.gitignore',
                         'website/.gitignore',
                         'website/.gitignore',
                         'website/.gitignore',
                         'website/.gitignore',
                         'website/README.md',
                         'website/README.md',
                         'website/README.md',
                         'website/README.md',
                         'website/README.md',
                         'website/README.md',
                         'website/README.md',
                         'website/jsconfig.json',
                         'website/jsconfig.json',
                         'website/jsconfig.json',
                         'website/jsconfig.json',
                         'website/jsconfig.json',
                         'website/jsconfig.json',
                         'website/jsconfig.json',
                         'website/next.config.js',
                         'website/next.config.js',
                         'website/next.config.js',
                         'website/next.config.js',
                         'website/next.config.js',
                         'website/next.config.js',
                         'website/next.config.js',
                         'website/package.json',
                         'website/package.json',
                         'website/package.json',
                         'website/package.json',
                         'website/package.json',
                         'website/package.json',
                         'website/package.json',
                         'website/postcss.config.js',
                         'website/postcss.config.js',
                         'website/postcss.config.js',
                         'website/postcss.config.js',
                         'website/postcss.config.js',
                         'website/postcss.config.js',
                         'website/postcss.config.js',
                         'website/public/*',
                         'website/src/components/*',
                         'website/src/pages/*',
                         'website/src/styles/*',
                         'website/tailwind.config.js',
                         'website/tailwind.config.js',
                         'website/tailwind.config.js',
                         'website/tailwind.config.js',
                         'website/tailwind.config.js',
                         'website/tailwind.config.js',
                         'website/tailwind.config.js']}

install_requires = \
['Flask-Cors>=3.0.10,<4.0.0',
 'Flask-SocketIO>=5.3.4,<6.0.0',
 'attrs==21.4.0',
 'boto3>=1.26.142,<2.0.0',
 'certifi==2022.6.15',
 'charset-normalizer==2.1.0',
 'idna==3.3',
 'iniconfig==1.1.1',
 'langchain>=0.0.183,<0.0.184',
 'nats-py>=2.3.1,<3.0.0',
 'openai>=0.27.7,<0.28.0',
 'packaging==21.3',
 'pinecone-client>=2.2.1,<3.0.0',
 'pluggy==1.0.0',
 'py==1.11.0',
 'pyparsing==3.0.9',
 'pypdf>=3.9.0,<4.0.0',
 'requests-mock>=1.9.3,<2.0.0',
 'requests==2.28.1',
 'returns>=0.19.0,<0.20.0',
 'simplejson>=3.17.6,<4.0.0',
 'tiktoken>=0.4.0,<0.5.0',
 'tomli==2.0.1',
 'tqdm>=4.65.0,<5.0.0',
 'types-requests>=2.28.0,<3.0.0',
 'types-simplejson>=3.17.7,<4.0.0',
 'urllib3==1.26.9']

setup_kwargs = {
    'name': 'seaplane',
    'version': '0.3.19',
    'description': 'Seaplane Python SDK',
    'long_description': '# Seaplane Python SDK\n[![PyPI](https://badge.fury.io/py/seaplane.svg)](https://badge.fury.io/py/seaplane)\n[![Python](https://img.shields.io/pypi/pyversions/seaplane.svg?style=plastic)](https://badge.fury.io/py/seaplane)\n\nSimple Python library to manage your resources at seaplane.\n\n## What is Seaplane?\n\nSeaplane is the global platform for building and scaling your application stack\nwithout the complexity of managing cloud infrastructure.\n\nIt serves as a reference application for how our APIs can be utilized.\n\nNot sure where to go to quickly run a workload on Seaplane? See our [Getting\nStarted] guide.\n\nTo build and test this software yourself, see the CONTRIBUTING document that is a peer to this one.\n\n## Installation\n\n```shell\npip install seaplane\n```\n\n## Configure your API KEY\n\n* Set `SEAPLANE_API_KEY` environment variable.\n* Use `config` object in order to set the api key.\n\n```python\nfrom seaplane import sea\n\nsea.config.set_api_key("your_api_key")\n```\n\n## License\n\nLicensed under the Apache License, Version 2.0, [LICENSE]. Copyright 2022 Seaplane IO, Inc.\n\n[//]: # (Links)\n\n[Seaplane]: https://seaplane.io/\n[CLI]: https://github.com/seaplane-io/seaplane/tree/main/seaplane-cli\n[SDK]: https://github.com/seaplane-io/seaplane/tree/main/seaplane\n[Getting Started]: https://github.com/seaplane-io/seaplane/blob/main/seaplane-sdk/python/docs/quickstart.md\n[CONTRIBUTING]: https://github.com/seaplane-io/seaplane/tree/main/seaplane-sdk/python/CONTRIBUTIONS.md\n[LICENSE]: https://github.com/seaplane-io/seaplane/blob/main/LICENSE\n',
    'author': 'Seaplane IO, Inc.',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/seaplane-io/seaplane/tree/main/seaplane-sdk/python',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Callable, Dict, Optional, Tuple

import requests

from ...configuration import config
from ...logging import log
from ...model.errors import SeaplaneError

REPLICATE_URL = "https://api.replicate.com/v1"
_REPLICATE_API_KEY_NAME = "REPLICATE_API_KEY"


def _check_replicate_api_key() -> None:
    if config._api_keys is None:
        raise SeaplaneError(
            f"Replicate API Key `{_REPLICATE_API_KEY_NAME}` is not set,\
                  use `sea.config.set_api_keys`."
        )
    elif not config._api_keys.get(_REPLICATE_API_KEY_NAME, None):
        raise SeaplaneError(
            f"Replicate API Key `{_REPLICATE_API_KEY_NAME}` is not set,\
                  use `sea.config.set_api_keys`."
        )


def inference(version: str) -> Callable[[Dict[str, Any]], Any]:
    def model(input: Dict[str, Any]) -> Any:
        """
        Replicate input example for stable diffusion:

            "input": {
                "width": 512,
                "height": 512,
                "num_outputs": 4,
                "prompt": "prompt",
                "seed": 11,
            },
        """
        _check_replicate_api_key()

        headers = {"Authorization": f"Token {config._api_keys[_REPLICATE_API_KEY_NAME]}"}

        payload = {"version": version, "input": input}

        response = requests.post(
            f"{REPLICATE_URL}/predictions",
            json=payload,
            headers=headers,
        )

        if response.ok:
            result = response.json()
            id = result["id"]

            while not result.get("output", None):
                endpoint = f"{REPLICATE_URL}/predictions/{id}"
                response = requests.get(
                    endpoint,
                    headers=headers,
                )

                log.info("Loading...")
                if response.ok:
                    result = response.json()

                    if result.get("output", None) is not None:
                        return result
        else:
            return response.text

    return model


class Replicate:
    def __init__(
        self, func: Callable[[Any], Any], type: str, id: str, model: Optional[str]
    ) -> None:
        self.func = func
        self.args: Optional[Tuple[Any, ...]] = None
        self.kwargs: Optional[Dict[str, Any]] = None
        self.type = type
        self.model = model
        self.id = id

    def process(self, *args: Any, **kwargs: Any) -> Any:
        self.args = args
        self.kwargs = kwargs

        if self.model == "stable-diffusion":
            log.info("Accessing Replicate stable-diffusion coprocessor...")
            self.args = self.args + (
                inference("27b93a2413e7f36cd83da926f3656280b2931564ff050bf9575f1fdf9bcd7478"),
            )

            return self.func(*self.args, **self.kwargs)
        elif self.model:
            log.info(f"Accessing Replicate {self.model} coprocessor...")
            self.args = self.args + (inference(self.model),)

            return self.func(*self.args, **self.kwargs)

    def print(self) -> None:
        log.info(f"id: {self.id}, type: {self.type}, model: {self.model}")
